;
; STE PLAYFIELD ROUTINES
; allow GFABASIC to have access to an independent playfield screen
; with STE hardware scrolling and a split screen (panel) with
; different palette
;
; (c) 2007/2008 by Simon Sunnyboy / Paradize
; based upon Scrolly STE Screen technology
;
; Special acknowledge to Klapauzius and Nicolas Pomarede for the split stabilizer!
;
; USE AT YOUR OWN RISK!
;
; STE/F030 only!
;
EXPORT STEPlayfield_Install
EXPORT STEPlayfield_Uninstall
EXPORT STEPlayfield_SplitLocation
EXPORT STEPlayfield_UpperScreen
EXPORT STEPlayfield_LowerScreen

; ** register equates **

videobase_h     EQU $FFFF8201
videobase_m     EQU $FFFF8203
videobase_l     EQU $FFFF820D
videocount_h    EQU $FFFF8205
videocount_m    EQU $FFFF8207
videocount_l    EQU $FFFF8209
linewidth       EQU $FFFF820F
hscroll         EQU $FFFF8265
palette         EQU $FFFF8240

iera            EQU $FFFFFA07
ierb            EQU $FFFFFA09
ipra            EQU $FFFFFA0B
iprb            EQU $FFFFFA0D
imra            EQU $FFFFFA13
imrb            EQU $FFFFFA15
isra            EQU $FFFFFA0F
isrb            EQU $FFFFFA11
tbdr            EQU $FFFFFA21
tbcr            EQU $FFFFFA1B

count_scanlines EQU 8
; ** TOS system variables **
_vbclock        EQU $0462
_frclock        EQU $0466

; ** ISR locations ***
vbl             EQU $70
timerb          EQU $0120
ikbd            EQU $0118

; ** datastructure offsets **
scradr_h        EQU 1
scradr_m        EQU 2
scradr_l        EQU 3
lwidth          EQU 36
hscroll_data    EQU 37
colors          EQU 4

				TEXT
				SUPER

                DC.B "STE PLAYFIELD KERNEL by Paradize v1.0 (AHCC version)",0
                DC.B "code by Simon Sunnyboy <marndt@asmsoftware.de>",0
                DC.B "special thanks to Klapauzius and Nicolas Pomarede for the Timer B stabilization",0
                EVEN

; ** init split screen **
STEPlayfield_Install:
                movem.l D0-A6,-(SP)
; setup split location
				move.b d0,STEPlayfield_SplitLocation
; store interrupt context
                lea     storage,A3
                move.l  vbl,(A3)+       ; save old VBL
                move.l  timerb,(A3)+    ; save old Timer B
                move.b  tbcr,(A3)+
                move.b  tbdr,(A3)+
                move.b  iera,(A3)+
                move.b  ierb,(A3)+
                move.b  imra,(A3)+
                move.b  linewidth,(A3)+
                move.b  hscroll,(A3)+

                move    #$2700,SR       ; stop interrupts

                move.l  #playfield_vbl,vbl          ; set new VBL
                move.l  #split_timerb,timerb  		; set new Timer B

; enable Timer B
                ori.b   #1,iera         ; enable Timer B
                ori.b   #1,imra         ; mask Timer B
                clr.b   tbcr            ; stop Timer B

                move    #$2300,SR       ; reenable interrupts
                movem.l (SP)+,D0-A6
                rts

; ** disable splitscreen **
STEPlayfield_Uninstall:
                move.l  A3,-(SP)

                move    #$2700,SR       ; stop irq system

                lea     storage,A3
                move.l  (A3)+,vbl       ; restore old VBL
                move.l  (A3)+,timerb    ; save old Timer B
                move.b  (A3)+,tbcr      ; restore MFP context
                move.b  (A3)+,tbdr
                move.b  (A3)+,iera
                move.b  (A3)+,ierb
                move.b  (A3)+,imra
                move.b  (A3)+,linewidth
                move.b  (A3)+,hscroll

                move    #$2300,SR       ; reenable irqs

                movea.l (SP)+,A3
                rts

; ** playfield handling VBL **
playfield_vbl:
                movem.l D0-A3,-(SP)
                clr.b   tbcr            ; stop Timer B

; --- display playfield ---


; set video for upper screen

                move.b  STEPlayfield_UpperScreen+scradr_h,videobase_h ; set videobase for compatibility
                move.b  STEPlayfield_UpperScreen+scradr_m,videobase_m
                move.b  STEPlayfield_UpperScreen+scradr_l,videobase_l
                move.b  STEPlayfield_UpperScreen+hscroll_data,hscroll ; set scroll registers
                move.b  STEPlayfield_UpperScreen+lwidth,linewidth
; patch for STE Shifter bug: reinitialize internal shifter timing
;                            by writing videocount
                move.b  STEPlayfield_UpperScreen+scradr_h,videocount_h ; set videocount (actual display)
                move.b  STEPlayfield_UpperScreen+scradr_m,videocount_m
                move.b  STEPlayfield_UpperScreen+scradr_l,videocount_l
; configure screen split
                move.b  STEPlayfield_SplitLocation,tbdr
                move.b  #count_scanlines,tbcr
; set palette of main screen
                movem.l STEPlayfield_UpperScreen+colors,D0-D7
                movem.l D0-D7,palette

                movem.l (SP)+,D0-A3
; push old VBL address onto stack
                move.l  storage,-(SP)
                rts                     ; and jump to it :)

; ** screensplit handling Timer B ISR **
split_timerb:
				movem.l D0-D7/A0-A1,-(SP)
                clr.b   tbcr            ; stop Timer B

				; we have to wait for the new scaline so use the time to prepare the
				; data that needs to be written into HW registers with start of the new scanline
				lea	 videobase_h,A0
				lea	 STEPlayfield_LowerScreen,A1
				move.l colors(a1),d6    ; set color 0 and 1
				move.b scradr_h(a1),d1
				move.b scradr_m(a1),d2
				move.b scradr_l(a1),d3
				move.b hscroll_data(a1),d4
				move.b lwidth(a1),d5

				; wait for scanline end (Klapauzius  style)
				move.b 8(a0),d0   ; $ffff8209 == video address counter low byte
wait_for_start_of_line:
				cmp.b 8(a0),d0
				beq.s wait_for_start_of_line

wait_for_end_of_line:  move.b 8(a0),d0
				cmp.b 8(a0),d0
				bne.s wait_for_end_of_line


; set new video for panel at new scanline:

                move.b  d4,hscroll ; set scroll registers
                move.b  d5,linewidth

                move.b  d1,videocount_h ; set videocount (actual display)
                move.b  d2,videocount_m
                move.b  d3,videocount_l
				move.l  d6, palette

; set palette of split screen
                movem.l colors+4(a1),D1-D7  ; color 0 and 1 have been set already!
                movem.l D1-D7,palette+4


				movem.l (SP)+,D0-D7/A0-A1

                bclr    #0,isra         ; acknowledge Timer B

                rte

                BSS
storage:		DS.L 4
STEPlayfield_UpperScreen:
				DS.W 19
STEPlayfield_LowerScreen:
				DS.W 19
				EVEN
STEPlayfield_SplitLocation:
				DS.B 1
				EVEN
                END
